<?php
/**
 * @file
 * Contains helper classes for the SDL BeGlobal plugin specific operations
 * depending on the translation quality selected by the user.
 */

/**
 * Interface TMGMTSDLBeGlobalTranslatorPluginHelperInterface
 */
interface TMGMTSDLBeGlobalTranslatorPluginHelperInterface {
  function doPost($action, $parameters = array(), $tokens = array());
  function doGet($action, $parameters = array(), $tokens = array());
  function doPut($action, $parameters = array(), $tokens = array());
  function doDelete($action, $parameters = array(), $tokens = array());
  function doRequestTranslation(TMGMTJob $job);
}

/**
 * Class TMGMTSDLBeGlobalTranslatorPluginHelperAbstract
 */
abstract class TMGMTSDLBeGlobalTranslatorPluginHelperAbstract implements TMGMTSDLBeGlobalTranslatorPluginHelperInterface {

  /**
   * The plugin controller which uses this helper.
   *
   * @var TMGMTSDLBeGlobalTranslatorPluginController
   */
  protected $plugin_controller;

  /**
   * The translator object.
   *
   * @var TMGMTTranslator
   */
  protected $translator;

  /**
   * The action for which this helper is required. It is a uri form a list of
   * available actions validated by the controller.
   *
   * @var string
   */
  protected $action;

  /**
   * Constructs a new helper for the translation plugin.
   *
   * @param TMGMTSDLBeGlobalTranslatorPluginController $plugin_controller
   * @param TMGMTTranslator $translator
   * @param string $action
   */
  public function __construct(TMGMTSDLBeGlobalTranslatorPluginController $plugin_controller,
                              TMGMTTranslator $translator, $action) {

    $this->plugin_controller = $plugin_controller;
    $this->translator = $translator;
    $this->action = $action;
  }

  /**
   * Creates a POST request to SDL BeGlobal Translate service.
   *
   * @param string $action
   *   (Optional) Uri for the resource we need to access.
   * @param array $parameters
   *   (Optional) Additional data parameters to be passed into the request.
   * @param array $tokens
   *   (Optional) Tokens to be replaced in the action url pattern.
   *
   * @return array
   *   Decoded JSON response from SDL BeGlobal Translate service.
   *
   * @throws TMGMTSDLBeGlobalException
   *   - Unable to connect to SDL BeGlobal Translate service
   *   - Error returned by SDL BeGlobal Translate service
   */
  public function doPost($action = '', $parameters = array(), $tokens = array()) {
    if (!empty($action)) {
      $this->action = $action;
    }
    $options = array(
      'method' => 'POST',
      'headers' => array(
        'Authorization' => 'BeGlobal apiKey=' . $this->plugin_controller->api_key,
      ),
    );
    if (!empty($parameters['file'])) {
      $boundary = 'A0sFSD';
      $options['data'] = $this->multipartEncode($boundary, $parameters);
      $options['headers']['Content-type'] = 'multipart/form-data; boundary=' . $boundary;
    }
    else {
      $options['data'] = json_encode($parameters);
      $options['headers']['Content-type'] = 'application/json';
    }

    return $this->processRequest(
      url($this->plugin_controller->getTranslationUri() . $this->replaceActionTokens($tokens)),
      $options
    );
  }

  /**
   * Creates a GET request to SDL BeGlobal Translate service.
   *
   * @param string $action
   *   (Optional) Uri for the resource we need to access.
   * @param array $parameters
   *   (Optional) Additional query parameters to be passed into the request.
   * @param array $tokens
   *   (Optional) Tokens to be replaced in the action url pattern.
   *
   * @return array
   *   Decoded JSON response from SDL BeGlobal Translate service.
   *
   * @throws TMGMTSDLBeGlobalException
   *   - Unable to connect to SDL BeGlobal Translate service
   *   - Error returned by SDL BeGlobal Translate service
   */
  public function doGet($action = '', $parameters = array(), $tokens = array()) {
    if (!empty($action)) {
      // If the actions is to get the stored file ID we need another option
      // For further processing the request.
      $options['fileStorage'] = 0;
      if ($action == 'files/storedFileId/file') {
        $options['fileStorage'] = 1;
      }
      $this->action = $action;
    }
    $options['headers']['Authorization'] = 'BeGlobal apiKey=' . $this->plugin_controller->api_key;

    return $this->processRequest(
      url(
        $this->plugin_controller->getTranslationUri() . '/' . $this->replaceActionTokens($tokens),
        array('query' => $parameters)
      ),
      $options
    );
  }

  /**
   * Creates a PUT request to SDL BeGlobal Translate service.
   *
   * @param string $action
   *   (Optional) Uri for the resource we need to access.
   * @param array $parameters
   *   (Optional) Additional query parameters to be passed into the request.
   * @param array $tokens
   *   (Optional) Tokens to be replaced in the action url pattern.
   *
   * @return array
   *   Decoded JSON response from SDL BeGlobal Translate service.
   *
   * @throws TMGMTSDLBeGlobalException
   *   - Unable to connect to SDL BeGlobal Translate service
   *   - Error returned by SDL BeGlobal Translate service
   */
  public function doPut($action = '', $parameters = array(), $tokens = array()) {
    if (!empty($action)) {
      $this->action = $action;
    }

    return $this->processRequest(
      url($this->plugin_controller->getTranslationUri() . '/' . $this->replaceActionTokens($tokens)),
      array(
        'data' => json_encode($parameters),
        'method' => 'PUT',
        'headers' => array(
          'Content-type' => 'application/json',
          'Authorization' => 'BeGlobal apiKey=' . $this->plugin_controller->api_key,
        ),
      )
    );
  }

  /**
   * Creates a DELETE request to SDL BeGlobal Translate service.
   *
   * @param string $action
   *   (Optional) Uri for the resource we need to access.
   * @param array $parameters
   *   (Optional) Additional query parameters to be passed into the request.
   * @param array $tokens
   *   (Optional) Tokens to be replaced in the action url pattern.
   *
   * @return array
   *   Decoded JSON response from SDL BeGlobal Translate service.
   *
   * @throws TMGMTSDLBeGlobalException
   *   - Unable to connect to SDL BeGlobal Translate service
   *   - Error returned by SDL BeGlobal Translate service
   */
  public function doDelete($action = '', $parameters = array(), $tokens = array()) {
    if (!empty($action)) {
      $this->action = $action;
    }
    $url = url($this->plugin_controller->getTranslationUri() . '/' . $this->replaceActionTokens($tokens));
    $options['headers']['Authorization'] = 'BeGlobal apiKey=' . $this->plugin_controller->api_key;

    return $this->processRequest($url, $options);
  }

  /**
   * Executes the request, validates the response and returns the decoded data.
   *
   * @param $url
   * @param $options
   *
   * @return bool|array
   */
  protected function processRequest($url, $options) {
    $response = drupal_http_request($url, $options);

    if (!$this->validateResponse($response)) {
      return FALSE;
    }

    // Return the processed JSON result.
    if (!empty($options['fileStorage'])) {
      return $response->data;
    }
    return drupal_json_decode($response->data);
  }

  /**
   * Validates the response received from SDL BeGlobal Translate service and
   * checks for thrown errors. This method should be extended in order to catch
   * the thrown exceptions in case there are any.
   *
   * @param object $response
   *   The response object from SDL BeGlobal Translate service.
   *
   * @return bool
   *
   * @throws TMGMTSDLBeGlobalException
   */
  protected function validateResponse($response) {
    if ($response->code != 200) {
      throw new TMGMTSDLBeGlobalException(
        t('Unable to connect to SDL BeGlobal Translate service due to following error: @error.'),
        array('@error' => $response->error)
      );
    }
    // If we got error.
    if (!empty($response->error)) {
      throw new TMGMTSDLBeGlobalException(
        t('SDL BeGlobal Translate service returned following error: @error'),
        array('@error' => $response->error)
      );
    }

    return TRUE;
  }

  /**
   * Checks if there are tokens in the url and take the replacement tokens from
   * the parameters.
   *
   * @param array $tokens
   *   An array containing as key the parameter (projectId, fileId, jobId) and
   *   as value the respective ID received from SDL BeGlobal.
   *
   * @return string
   *   The action with the tokens replaced
   */
  protected function replaceActionTokens($tokens = array()) {
    if (empty($tokens)) {
      return $this->action;
    }

    $this->action = str_replace(array_keys($tokens), array_values($tokens), $this->action);

    return $this->action;
  }

  /**
   * Helper function to process the parameters used for the file upload.
   *
   * @param string $boundary
   * @param array $parameters
   *
   * @return string
   */
  private function multipartEncode($boundary, $parameters) {
    $output = "--$boundary\r\n";
    foreach ($parameters as $key => $value){
      if ($key == 'file') {
        if (substr($value, 0, 1) == "@") {
          $value = substr($value, 1);
        }
        $filename = basename($value);
        $output .= "Content-Disposition: form-data; name=\"file\"; filename=\"$filename\"\r\n" .
          "Content-Transfer-Encoding: binary\r\n" .
          "Content-Type: application/octet-stream\r\n\r\n" .
          file_get_contents($value) . "\r\n";
      }
      else {
        $output .= "Content-Disposition: form-data; name=\"$key\"\r\n\r\n$value\r\n";
      }
      $output .= "--$boundary";
    }
    $output .= "--$boundary--";

    return $output;
  }
}

/**
 * Class TMGMTSDLBeGlobalTranslatorPluginHelperMachineTranslation
 */
class TMGMTSDLBeGlobalTranslatorPluginHelperMachineTranslation extends TMGMTSDLBeGlobalTranslatorPluginHelperAbstract{

  /**
   * {@inheritdoc}
   */
  protected function validateResponse($response) {
    try {
      parent::validateResponse($response);
      // If we do not have data.
      if (empty($response->data)) {
        throw new TMGMTSDLBeGlobalException(t('SDL BeGlobal Translate service returned no data'));
      }
    }
    catch (TMGMTSDLBeGlobalException $e) {
      drupal_set_message($e->getMessage(), 'error');

      return FALSE;
    }

    return TRUE;
  }

  /**
   * Helper function to deal with data processing in order to make a translation
   * request. It attempts to translate each value at a time. In case for one
   * value we have an error, we increase the error count and if none of the item
   * could be translated we reject the job.
   *
   * @param TMGMTJob $job
   */
  public function doRequestTranslation(TMGMTJob $job) {
    // Pull the source data array through the job and flatten it.
    $data = array_filter(tmgmt_flatten_data($job->getData()), '_tmgmt_filter_data');

    $translation = array();
    // Errors counter.
    $er = 0;

    // The keys with which the texts needs to be mapped.
    $keys = array_keys($data);

    // All of the texts to be translated.
    $values = array();

    foreach ($data as $key => $value) {
      array_push($values, $value['#text']);
    }

      try {
        $result = $this->doPost($this->action, array(
          'from' => $this->plugin_controller->mapToRemoteLanguage($job->getTranslator(), $job->source_language),
          'to' => $this->plugin_controller->mapToRemoteLanguage($job->getTranslator(), $job->target_language),
          'texts' => $values,
          // InputFormat (optional).
        ));

        if (empty($result['translations'])) {
          throw new TMGMTSDLBeGlobalException('Empty translation response string.');
        }

        // Combine the keys with the translation values.
        $i = 0;
        foreach($keys as $key) {
          $translation[$key]['#text'] = $result['translations'][$i];
          $i++;
        }
      }
      catch (TMGMTSDLBeGlobalException $e) {
        $job->addMessage('There was a translation error: !error for translation of text "!text."',
          array('!error' => $e->getMessage(), '!text' => $data[$key]['#text']), 'error');
        $er++;
      }

    if ($er != count($data)) {
      // The translation job has been successfully submitted.
      $job->submitted('The translation job @job_name has been submitted.', array('@job_name' => $job->label));
      // Save the translated data through the job.
      $job->addTranslatedData(tmgmt_unflatten_data($translation));
    }
    else {
      $job->rejected('None of the items could be translated.');
    }
  }
}

/**
 * Class TMGMTSDLBeGlobalTranslatorPluginHelperHumanTranslation
 */
class TMGMTSDLBeGlobalTranslatorPluginHelperHumanTranslation extends TMGMTSDLBeGlobalTranslatorPluginHelperAbstract {
  /**
   * Follows the steps to initiate a human translation from SDL BeGlobal with
   * these steps:
   * - each job item will be put in a .txt file because SDL BeGlobal do not
   * support csv files
   * - these files will become files attached to the SDL project by uploading
   * them to SDL.
   * - the response from the file creation will be mapped to a remote
   * mapping entity TMGMTRemote which will contain key items from the
   * response, as fileId, projectId in order to be able to process it later
   * when the callback is accessed for notification
   * - place the order for the project
   *
   * @param TMGMTJob $job
   *
   * @return bool
   */
  public function doRequestTranslation(TMGMTJob $job) {

    $job_items = $job->getItems();
    if (empty($job_items)) {
      return FALSE;
    }
    $project = $this->doPost('projects', array(
      'type' => 'translationProject',
      'from' => $this->plugin_controller->mapToRemoteLanguage($job->getTranslator(), $job->source_language),
      'to' => $this->plugin_controller->mapToRemoteLanguage($job->getTranslator(), $job->target_language),
      'name' => $job->label,
      'description' => $job->label,
      'notificationCallbackURL' => url('tmgmt_sdl_beglobal/tmgmt_sdl_beglobal_human_callback', array('absolute' => TRUE)),
    ));
    try{
      $file_ids = array();
      foreach ($job_items as $item) {
        $data = array_filter(tmgmt_flatten_data($item->data), '_tmgmt_filter_data');
        // Generate the xml path.
        $file_path_xml = 'public://sdl_beglobal/job_item_' . $item->tjiid . '.xml';
        // Upload a generated file for this job item.
        $file = $this->uploadFile($file_path_xml, $data, $project);
        // Store the ids of the files for later usage.
        array_push($file_ids, $file['storedFiles'][0]['id']);
        // Create the remote mapping entity for the current job item.
        $remote_mapping = $item->addRemoteMapping(
          serialize(array_keys($data)),
          $project['project']['id'],
          array(
            'remote_identifier_2' => $file['storedFiles'][0]['id'],
            'remote_data' => serialize($file),
          )
        );
        if (!$remote_mapping) {
          throw new TMGMTSDLBeGlobalException('The remote mapping could not be created.');
        }
      }

      // Add files to the project.
      $result = $this->doPut('projects/projectId/files',
        array('fileIds' => $file_ids),
        array('projectId' => $project['project']['id'])
      );

      if($result) {
        // Make the order to SDL BeGlobal. We override the old value of the project
        // because this call should bring new data into the project.
        $project = $this->doPut('projects/projectId/orders',
          array('qualityLevel' => $this->plugin_controller->getTranslationQuality()),
          array('projectId' => $project['project']['id'])
        );

        // Save a remote mapping for the job itself which corresponds to the sdl
        // beglobal project.
        $job = $this->addJobRemoteMapping($job, array(
          'remote_identifier_1' => $project['project']['id'],
          'remote_data' => serialize($project),
        ));
        $job->submitted('Job @job_name has been submitted to translator.',
          array('@job_name' => $job->label));
      }
    }
    catch (TMGMTSDLBeGlobalException $e) {
      // Reject the current job.
      $job->rejected($e->getMessage());
      // TODO cancel remote order also.
      // Delete the newly created project.
      $this->doDelete('/projects/projectId', array(),
        array('projectId' => $project['project']['id']));
      watchdog('TMGMT', $e->getMessage(), NULL, WATCHDOG_ERROR, NULL);

      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function validateResponse($response) {
    try {
      parent::validateResponse($response);
    }
    catch (TMGMTSDLBeGlobalException $e) {
      drupal_set_message($e->getMessage(), 'error');
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Uploads a file to SDL BeGlobal and attaches it to a project if provided.
   *
   * @param string $file_path
   * @param array $data
   * @param array $project
   *
   * @return array
   *
   * @throws TMGMTSDLBeGlobalException
   *   - In case of empty data object received from SDL BeGlobal.
   */
  private function uploadFile($file_path_xml, $data, $project = array()) {
    if (empty($data)) {
      throw new TMGMTSDLBeGlobalException(
        'Empty data object received for file @file',
        array('@file' => $file_path_xml)
      );
    }

    $xml = $this->generateXML($data);

    $dir_name = dirname($file_path_xml);
    if (file_prepare_directory($dir_name, FILE_CREATE_DIRECTORY)) {
      file_save_data($xml, $file_path_xml, FILE_EXISTS_REPLACE);
      $parameters['file'] = file_create_url($file_path_xml);
      if (!empty($project)) {
        $parameters += array(
          'projectId' => $project['project']['id'],
          'token' => $project['uploadToken'],
        );
      }
    }
    else {
      throw new TMGMTSDLBeGlobalException(
        'Please check that your /sites/default/files privileges
         are set to a user/group set created specifically for web servers
         in order to create the directory',
        array('@file' => $dir_name)
      );
    }

    return $this->doPost('files', $parameters);
  }

  /**
   * Creates a remote mapping entity for the remote sdl beglobal project.
   *
   * @param TMGMTJob $job
   * @param array $mapping_data
   *
   * @return mixed
   *
   * @throws TMGMTSDLBeGlobalException
   */
  private function addJobRemoteMapping(TMGMTJob $job, $mapping_data = array()) {
    if (empty($mapping_data['remote_identifier_1'])) {
      throw new TMGMTSDLBeGlobalException('Cannot create remote mapping without remote identifier.');
    }

    // We map only the job on this remote mapper.
    $data = array(
      'tjid' => $job->tjid,
      'tjiid' => 0,
    );

    if (!empty($mapping_data)) {
      $data += $mapping_data;
    }

    $remote_mapping = entity_create('tmgmt_remote', $data);
    if ($x = entity_get_controller('tmgmt_remote')->save($remote_mapping)) {
      $job->reference = $remote_mapping->trid;
    }

    return $job;
  }

  /**
   * Generate the xml file to be uploaded.
   *
   * @param array $data
   *
   * @return string
   *   xml file formatted
   */
  private function generateXML($data) {
    // Get the corresponding field keys.
    if (empty($data)) {
      return '';
    }

    $keys = str_replace('][', '-', array_keys($data));
    // Get the corresponding field values.
    $values = array();
    foreach ($data as $value) {
      array_push($values, $value['#text']);
    }
    // Combine the keys->values.
    $combined_array = array_combine($keys, $values);

    // Generate the xml format.
    $xml_format = '<?xml version="1.0" encoding="UTF-8"?>' . "\n" . '<document>' . "\n" . format_xml_elements($combined_array) . '</document>';

    return $xml_format;
  }
}

